class Describe:

    def __init__(self, tokens):

        '''Takes in a list with tokens.'''

        from signs.preprocess import lists_to_list

        if isinstance(tokens[0], str):
            self.tokens = tokens
        if isinstance(tokens[0], list):
            self.tokens = lists_to_list(tokens)

    def get_counts(self):

        '''Get word frequencies for the aggregate
        of all the tokens in Describe() input.'''

        from collections import Counter

        # return the counted values as dictionary
        return dict(Counter(self.tokens).most_common())

    def get_gram_counts(self, ngram=3, skip=0):

        '''Returns ngram counts. Supports skipgrams.

        docs : list
            A list of lists where each sublist contains a string.
        ngram : int
            The value for 'n' in ngram. For example, 2 for bigram.
        skip : int
            The value for skipgrams. For example, 3 to skip three tokens
        '''

        from collections import Counter
        from signs import Grams, preprocess

        temp = [Grams(self.tokens).ngrams(ngram, skip)]
        temp = preprocess.lists_to_list(temp)
        temp = [' '.join(i) for i in temp]

        return dict(Counter(temp).most_common())



def similarity(self, word1, word2):

    '''WORD SIMILARITY
    Takes as input two words and returns similarity.

    PARAMS
    ------
    word1 :: a string
    word2 :: another string


    EXAMPLE
    -------
    get = Signs(backend='gensim',
    trained_vectors="e:/signs_dev/vectors/glove/glove.twitter.27B.25d.txt")
    get.similarity('donald', 'trump')

    '''

    if self.backend == 'gensim':
        return self.model.similarity(word1, word2)

    if self.backend == 'spacy':

        a = self.model(word1)
        b = self.model(word2)

        return a.similarity(b)



def gensim_doc2vec_train(docs):
    
    '''Trains a gensim doc2vec model based on a training corpus. 
    Returns the trained model and the training docs.
    
    NOTE: the input docs format is list-of-lists where each
    sublists consist of tokenized document.

    EXAMPLE:

        model, train_corpus = gensim_doc2vec_train(docs[:400])
        test_corpus = docs[400:]

    '''

    from gensim.models.doc2vec import Doc2Vec, TaggedDocument

    documents = [TaggedDocument(doc, [i]) for i, doc in enumerate(docs)]
    
    model = Doc2Vec(vector_size=50, 
                    window=2,
                    min_count=2,
                    workers=4,
                    epochs=50)
    
    model.build_vocab(documents)
    
    model.train(documents,
                total_examples=model.corpus_count,
                epochs=model.epochs)
    
    return model, documents


class AutoPrep:

    def __init__(self, docs):

        self.docs = self._format(docs)
        self.docs = self._clean()

    def _format(self, docs):

        # input is a single string
        if isinstance(docs, str):
            pass
        # input is list with strings
        if isinstance(docs[0], str):
            pass
        # input is list with lists
        if isinstance(docs[0], list):
            pass

    def _clean(self):

        from signs import Clean
        return [[Clean(doc).text] for doc in self.docs]



import spacy as sp


def spacy_load(trained_vectors='en'):

    '''LOAD SPACY MODEL

    The default option 'en' loads the spacy default vectors.
    Alternative a local vector file can be loaded or one of the
    other spacy vectors.

    '''

    if trained_vectors == 'en':

        import en_core_web_sm
        return en_core_web_sm.load()

    else:
        return sp.load(trained_vectors)
    
    
    


def spacy_word2vec(string, model):

    '''GET VECTOR REPRESENTATION

    Takes as input a string (word or sentence) and
    returns the vector representation based on trained vectors.

    '''

    return model(string).vector



import pandas as pd
from ..utils.converters import embeds_to_text
try:
    from kerasplotlib.text import text
except ImportError:
    print('Matplotlib failed to import')


class Preds:

    def __init__(self, x_test, y_test, word_index, model):

        # temp values
        self.model = model
        self.x_test = x_test
        self.y_test = y_test
        self.word_index = word_index

        self.results = self._preds_df()

        # delete temp values
        del self.word_index, self.y_test, self.x_test, self.model

    def _preds_df(self):

        '''
        x_test :: the test data (already embedded)
        y_test :: truth values for the test data
        word_index :: the word index used for creating the embeddings

        '''

        results = pd.DataFrame({
                    'text': embeds_to_text(self.x_test, self.word_index),
                    'pred': [i[0] for i in self.model.predict(self.x_test)],
                    'truth': self.y_test
                            })

        return results

    def _printing(self, data, title, n, notebook):

        data.drop_duplicates(inplace=True)

        if notebook:
            text(data, 'text', title=title)
        else:
            print(title)
            for t in data['text'].head(n):
                print(t)
            print('\n')


    def summary(self, sensitivity=.1, n=5, notebook=True):

        '''Provides a summary of the classification results for the model

        sensitivity : float
             Cut-off point for deciding between positive and negative.
        n : int
             Number of samples to display.
        notebook : bool
             Use html printing for notebooks.

        '''

        pos = self.results[self.results.pred > 1 - sensitivity].sample(n, replace=True)
        self._printing(pos, 'CLEAR POSITIVE', n, notebook)

        mid = self.results[self.results.pred.between(0.5 - (sensitivity / 2), 0.5 + (sensitivity / 2))].sample(n, replace=True)
        self._printing(mid, 'CLOSE CALL', n, notebook)

        neg = self.results[self.results.pred < sensitivity].sample(n, replace=True)
        self._printing(neg, 'CLEAR NEGATIVE', n, notebook)


    def misses(self, sensitivity=.5, n=5, notebook=True):

        '''Random examples of false positives and false negatives

        sensitivity : float
             Cut-off point for deciding between positive and negative.
        n : int
             Number of samples to display.
        notebook : bool
             Use html printing for notebooks.

        '''

        # false positive
        fp = self.results[self.results.truth == 0][self.results.pred > 1 - sensitivity].sample(n, replace=True)
        self._printing(fp, 'FALSE POSITIVE', n, notebook)

        # false negatives
        fn = self.results[self.results.truth == 1][self.results.pred < sensitivity].sample(n, replace=True)
        self._printing(fn, 'FALSE NEGATIVE', n, notebook)

    def hits(self, sensitivity=.1, n=5, notebook=True):

        '''Random examples of true positives and true negatives

        sensitivity : float
             Cut-off point for deciding between positive and negative.
        n : int
             Number of samples to display.
        notebook : bool
             Use html printing for notebooks.

        '''

        # true positive
        tp = self.results[self.results.truth == 1][self.results.pred > 1 - sensitivity].sample(n, replace=True)
        self._printing(tp, 'TRUE POSITIVE', n, notebook)

        # true negatives
        tn = self.results[self.results.truth == 0][self.results.pred < sensitivity].sample(n, replace=True)
        self._printing(tn, 'TRUE NEGATIVE', n, notebook)



def lists_to_list(list_of_lists):

    '''Takes in list of lists and makes it into a flat list'''

    return [item for sublist in list_of_lists for item in sublist]


def strings_to_tokens(docs):

    '''Converts list-of-lists where each sublist
    is a string, to list-of-lists where each sublist
    consist of tokens'''

    return [doc[0].split() for doc in docs if doc[0] is not None]


def docs_to_list(docs):

    '''convert list of sentences to list of word'''

    return [word for line in docs for word in line.split()]


def docs_to_blob(docs):

    '''takes in list of words or sentences
    and returns a single blob of text.'''

    return ' '.join(word for word in docs)


def lists_to_blob(docs):

    return ' '.join([''.join(doc) for doc in docs if doc is not None])


def create_tokens(docs, flatten=True, clean=True):

    '''Takes in a list-of-lists where each sublist
    contains of a string. For example, a paragraph.

    Returns a single list where strings have
    been converted into tokens.'''

    from signs import Clean

    if clean:
        docs = [[Clean(doc, auto=True).text] for doc in docs]

    tokens = strings_to_tokens(docs)

    if flatten:
        return lists_to_list(tokens)
    else:
        return tokens



import numpy as np


def embedding_matrix(word_index,
                     vector_dims,
                     embeddings_index,
                     max_num_words=None):

    if max_num_words is None:
        max_num_words = len(word_index) + 1

    num_words = min(max_num_words, len(word_index) + 1)
    embedding_matrix = np.zeros((num_words, vector_dims))

    for word, i in word_index.items():
        if i >= num_words:
            continue
        embedding_vector = embeddings_index.get(word)

        if embedding_vector is not None:
            embedding_matrix[i] = embedding_vector

    return embedding_matrix



def read_file(filename):

    '''Takes as input a file where each row will become
    a document in the output list

    filename : str
        Local file with content

    '''

    with open(filename) as f:
        content = f.readlines()

    content = [x for x in content]

    return content



# coding=utf-8

def stopwords():

    nltk_stop = ['i', 'me', 'my', 'myself', 'we', 'our', 'ours', 'ourselves', 'you', 'your', 'yours', 'yourself', 'yourselves', 'he', 'him', 'his', 'himself', 'she', 'her', 'hers', 'herself', 'it', 'its', 'itself', 'they', 'them', 'their', 'theirs', 'themselves', 'what', 'which', 'who', 'whom', 'this', 'that', 'these', 'those', 'am', 'is', 'are', 'was', 'were', 'be', 'been', 'being', 'have', 'has', 'had', 'having', 'do', 'does', 'did', 'doing', 'a', 'an', 'the', 'and', 'but', 'if', 'or', 'because', 'as', 'until', 'while', 'of', 'at', 'by', 'for', 'with', 'about', 'against', 'between', 'into', 'through', 'during', 'before', 'after', 'above', 'below', 'to', 'from', 'up', 'down', 'in', 'out', 'on', 'off', 'over', 'under', 'again', 'further', 'then', 'once', 'here', 'there', 'when', 'where', 'why', 'how', 'all', 'any', 'both', 'each', 'few', 'more', 'most', 'other', 'some', 'such', 'no', 'nor', 'not', 'only', 'own', 'same', 'so', 'than', 'too', 'very', 's', 't', 'can', 'will', 'just', 'don', 'should', 'now', 'd', 'll', 'm', 'o', 're', 've', 'y', 'ain', 'aren', 'couldn', 'didn', 'doesn', 'hadn', 'hasn', 'haven', 'isn', 'ma', 'mightn', 'mustn', 'needn', 'shan', 'shouldn', 'wasn', 'weren', 'won', 'wouldn']
    twitter_stop = ['RT', 'http', 'https', 'rt', 'via']
    generic_stop = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z','1','2','3','4','5','6','7','8','9','0','was','are','going','used','&amp;','has','dont','amp',     'the','be','and','of','a','in','to','is','i','an',     'have','to','it','I','that','for','you','why','were',     'he','with','on','do','say','this','they','had','been',     'at','but','we','his','from','that','not','\"the',     'by','she','or','as','what','go','their','did',     'can','who','get','if','would','her','all','you.',     'my','make','about','know','will','as','up','let',     'one','time','there','so','when','which','them','did',     'some','me','take','out','into','just','see','him',     'your','come','could','now','than','like','other',     'how','then','its','our','two','more','these','don\'t',     'want','way','look','first','also','new','because',     'day','more','use','no','man','find','here','thing',     'give','many','well','only','those','tell','one',     'very','her','even','back','any','good','through',     'us','there','down','may','should','over','still',     'try','in','as','last','ask','need','too','feel',     'three','when','never','become','between','high','each',     'really','something','most','another','much','own','both',     'out','leave','put','old','while','mean','it\'s',     'im','u','i\'m','said','de','http','https','got',     'didn\'t','doesnt','didnt\…','i\'d','can\'t','doesn\'t',     'isn\'t','\“what\’s','ever','again.','()','thought',     'before','after','2016','knows','everyone','every','please']

    return nltk_stop + twitter_stop + generic_stop


class Stopwords:

    def __init__(self,
                 docs,
                 common_stopwords=True,
                 add_stopwords=[],
                 min_length=2,
                 max_threshold=10):

        '''Accepts as input a list-of-lists where
        each sublist is a document represented in tokens.

        docs : list (of lists)
            Tokenized documents.
        common_stopwords : bool
            If a comprehensive list of stopwords should be used. If set
            to False then add_stopwords can't be empty.
        add_stopwords : list or None
            If a list of words is provided, then those will be used
            as well as common_stopwords unless it's set to False.
        min_lenght : int
            Drop all words below this length.
        max_threshold : int
            Keep all words that are at least this long.

        '''

        import string

        self.common_stopwords = common_stopwords
        self.add_stopwords = add_stopwords
        self.min_length = min_length
        self.max_threshold = max_threshold

        self.string = string
        self.stopwords = self.stopword_index()
        self.docs = docs

        for i in range(len(docs)):
            self.docs[i] = self.check_stopwords(docs[i])

    def stopword_index(self):

        import numpy as np

        out = []

        if self.common_stopwords is True:
            from signs.utils.stopwords import stopwords
            stopword_list = np.unique(stopwords()).tolist()
            stopword_list + self.add_stopwords
        else:
            stopword_list = self.add_stopwords

        for word in stopword_list:
            if len(word) > 1:
                out.append(word.lower())

        stopword_dict = {}

        for word in out:

            if word[0] in self.string.ascii_letters:
                try:
                    stopword_dict[word[0]].append(word)
                except KeyError:
                    stopword_dict[word[0]] = [word]

        return stopword_dict

    def check_stopwords(self, doc):

        out = []

        for word in doc:

            # always keep words longer than 10 characters
            if len(word) >= self.max_threshold:
                out.append(word)

            # always pass words shorter than 2 characters
            elif len(word) <= self.min_length:
                continue

            elif word[0] not in self.stopwords.keys():
                out.append(word)

            elif word not in self.stopwords[word[0]]:
                out.append(word)

        return out



class Verbatims:

    def __init__(self, text):

        self.text = text
        self.counter = len(text)

    def verbatims(self, keyword, no_of_words=1):

        # the output
        out = []

        # process the doc word-by-word
        for i in range(self.counter):
            try:
                if self.text[i] == keyword:
                    words = self.text[i-no_of_words:i+1+no_of_words]
                    out.append(' '.join(words))
            except IndexError:
                pass

        return out



import os

import spacy as sp
try:
    import en_core_web_sm
except ModuleNotFoundError:
    pass

from gensim.models import KeyedVectors, FastText
from gensim.scripts.glove2word2vec import glove2word2vec


class LoadVectors:

    '''Load, train, and

    The default option 'en' loads the spacy default vectors.
    Alternative a local vector file can be loaded or one of the
    other spacy vectors.

    trained_vectors :: name or path to the pre-trained vector file. 
                       If set to 'en' then spacy small web vectors 
                       will be used.
    backend :: either 'spacy' or 'gensim'
    mode :: applies to backend 'gensim'. The options are 'glove', 'word2vec',
            and 'fasttext'
    temp :: filename for glove to word2vec conversion
    binary :: if the vector file format is binary or not

    '''

    def __init__(self,
                 trained_vectors='en',
                 backend='spacy',
                 mode=None,
                 temp='signs_temp.sig',
                 binary=False):

        self.backend = backend
        self.mode = mode
        self.temp = temp
        self.binary = binary
        self.trained_vectors = trained_vectors
        self._null = self.load_model()

    def load_model(self):

        if self.backend == 'spacy':

            if self.trained_vectors == 'en':

                import en_core_web_sm
                self.model = en_core_web_sm.load()
            
            else:
                self.model = sp.load(self.trained_vectors)

        elif self.backend == 'gensim':

            if self.mode is None:
                self.mode = 'glove'

            if self.mode == 'glove':
                glove2word2vec(self.trained_vectors, self.temp)
                self.model = KeyedVectors.load_word2vec_format(self.temp,
                                                               binary=self.binary)
                os.remove(self.temp)

            elif self.mode == 'word2vec':
                self.model = KeyedVectors.load_word2vec_format(self.trained_vectors,
                                                               binary=self.binary,
                                                               encoding='latin-1')
            elif self.mode == 'fasttext':
                self.model = FastText.load_fasttext_format(self.trained_vectors,
                                                           encoding='latin-1')

    def vectors(self, word):

        '''GET VECTOR REPRESENTATION

        Takes as input a string (word or sentence) and
        returns the vector representation based on trained vectors.

        '''

        if self.backend == 'spacy':
            return self.model(word).vector

        elif self.backend == 'gensim':
            return self.model[word]



def html_print(text, title=''):

    from IPython.core.display import display, HTML

    # create title for the content
    display(HTML("<h4>" + str(title) + "</h4>"))

    # create content
    html = display(HTML("<font size=2 face=Verdana>" + text + "</font>"))

    return html



import warnings
warnings.simplefilter('ignore')

from .commands.embeds import Embeds
from .commands.clean import Clean
from .commands.preds import Preds
from .commands.load_vectors import LoadVectors
from .commands.verbatims import Verbatims
from .commands.describe import Describe
from .commands.transform import Transform
from .commands.docsimilarity import DocSimilarity
from .commands.stopwords import Stopwords
from .commands.autoprep import AutoPrep

from .grams.grams import Grams
from .vectorize.gensim_doc2vec_train import gensim_doc2vec_train as TrainDoc2Vec
import signs.preprocess

__all__ = ['Embeds', 'Clean', 'Preds', 'LoadVectors', 'Grams', 'Describe', 'Transform']

del commands, utils, grams, vectorize, signs



import numpy as np


def embedding_index(trained_vectors):

    '''KERAS EMBEDDING INDEX

    Takes in trained vectors (e.g. GloVe) and returns
    a Keras embeddings index.

    PARAMS
    ------

    trained_vectors :: a file with the trained vectors

    '''

    embeddings_index = {}

    with open(trained_vectors, encoding='utf-8') as f:

        for line in f:

            values = line.split()
            word = values[0]
            try:
                coefs = np.asarray(values[1:], dtype='float32')
            except ValueError:
                pass
            embeddings_index[word] = coefs

    vector_dims = len(list(embeddings_index.values())[0])

    return embeddings_index, vector_dims



class Clean:

    def __init__(self, text, auto=True, remove_string=''):

        '''Allows several document cleaning workflows either
        individually, or as an automated pipeline. Accepts
        string value or list with single string value.'''

        # a custom string to be removed
        self._remove_string = remove_string

        # input is a string inside a list
        if isinstance(text, list):
            self.text = text[0]
        else:
            self.text = text

        if auto:
            self.automated()

    def automated(self):

        self.text = self.nones()
        self.text = self.decod()
        self.text = self.low()
        self.text = self.links()
        self.text = self.emoji()
        self.text = self.punct()
        self.text = self.linebreaks()
        self.text = self.string()
        self.text = self.leadtrail()
        self.text = self.whitespace()

    def nones(self, replace_with='NA'):

        '''Accepts as input a list-of-lists where each
        sublist contains a string. Replaces None values
        with an arbitrary string.'''

        if self.text is None:
            return replace_with
        else:
            return self.text

    def string(self, remove_string=None):

        '''remove arbitrary string'''

        if remove_string is None:
            remove_string = self._remove_string

        return self.text.replace(remove_string, '')

    def whitespace(self):

        '''Remove extra whitespaces'''

        return ' '.join(self.text.split())

    def low(self):

        '''make string lowercase'''

        return self.text.lower()

    def caps(self):

        '''make string uppercase'''

        return self.text.upper()

    def punct(self):

        '''remove special characters'''

        import re
        import string

        return re.sub('['+string.punctuation+']', '', self.text)

    def leadtrail(self):

        '''remove trailing and leading whitespace

        NOTE: this also removes the last line break'''

        return self.text.strip()

    def linebreaks(self):

        '''remove linebreaks'''

        return self.text.replace('\n', ' ').replace('\r', '')

    def decod(self):

        '''decode binary'''
        try:
            return self.text.decode()
        except AttributeError:
            return self.text

    def links(self):

        '''remove links'''

        import re

        return re.sub(r'http\S+', '', self.text)

    def emoji(self):

        '''remove emojis'''

        import re

        emoji_pattern = re.compile("["
                                   u"\U0001F600-\U0001F64F"
                                   u"\U0001F300-\U0001F5FF"
                                   u"\U0001F680-\U0001F6FF"
                                   u"\U0001F1E0-\U0001F1FF"
                                   u"\U00002702-\U000027B0"
                                   u"\U000024C2-\U0001F251"
                                   "]+", flags=re.UNICODE)
        return emoji_pattern.sub(r'', self.text)



class Grams:

    def __init__(self, text):

        '''Takes in a list of words representing a document
        and returns a list of ngrams or skipgrams.

        text : list
            A list of words representing a document.

        '''

        self._text = text
        self._counter = len(text)

    def ngrams(self, ngram=2, skip=0):

        '''Produce ngrams or skipgrams.

        ngram: int
            Number of words per gram.
        skip : int
            The number of words to skip for each gram'''

        # the output
        out = []

        # process the doc word-by-word
        for i in range(self._counter):
            try:
                # create the list for single instance of grams
                words = [self._text[i]]

                # find the grams
                for ii in range(ngram - 1):
                    words.append(self._text[i + ii + 1 + skip])

                # add to the output
                out.append(words)

            # handle the case where end is near
            except IndexError:
                pass

        return out

    def combinations(self, length=3):

        '''Produce every possible combination of n length.

        length | int | length of the combinations
        '''

        import itertools

        out = []

        for i in list(itertools.combinations(self._text, length)):
            if len(i) == len(set(i)):
                out.append(i)

        return out

    def flexgrams(self, length=3, flex_n=1):

        '''Simple flexgram where all combinations of certain length
        are returned, but n items are randomly dropped. The final
        length must be shorter than number of items minus flex_n

        length | int | length of the resulting combinations
        flex_n | int | the number of items to randomly drop

        '''
        import random
        import itertools

        combinations = list(itertools.combinations(self._text, length))

        out = []

        for i in [random.sample(i, k=length-flex_n) for i in combinations]:
            if len(i) == len(set(i)):
                if i not in out:
                    out.append(i)

        return out

    def sobolgrams(self, length=3, n=2):

        '''Sobolgram is a random variation of a flexgram, where
        instead of following a sequential order of strings and then
        randomly dropping words, there is no order at all and grams
        are formed based on random pick within the doc.

        length | int | length of the resulting combinations
        n | int | number of grams to draw

        '''

        import random

        out = []

        while len(out) < n:

            gram = random.sample(self._text, k=length)

            if len(set(gram)) == length:
                if gram not in out:
                    out.append(gram)

        return out

    def skipgrams(self, length=3, skip_nth=2):

        '''Simple flexgram where all combinations of certain length
        are returned, but n items are randomly dropped. The final
        length must be shorter than number of items minus flex_n

        length | int | length of the resulting combinations
        skip_nth | int | the number of item to skip

        '''

        import itertools
        import numpy as np

        combinations = list(itertools.combinations(self._text, length))

        # create indicies
        indicies = list(range(length))
        indicies.remove(skip_nth)

        out = []

        for i in [np.array(i)[indicies].tolist() for i in combinations]:
            if i not in out:
                out.append(i)

        return out



from ..preprocess.embedding_index import embedding_index
from ..preprocess.embedding_matrix import embedding_matrix
from ..preprocess.tokenize import tokenize

from keras.layers import Embedding


class Embeds:

    def __init__(self, trained_vectors):

        '''Takes as input trained vectors and returns an object which can be
        used to to create a Keras Embeds layer given input of documents.

        e = Embeds('glove.twitter.27B.25d.tx')
        embedding_layer  = e.layer(docs)

        '''

        self.t_v = trained_vectors
        self.embeddings_index, self.vector_dims = embedding_index(self.t_v)

    def _matrix(self):

        '''Helper function for processing input docs for the
        Keras Embeds layer'''

        self.x, self.vocab_size, self.word_index = tokenize(self.docs)
        self.embedding_matrix = embedding_matrix(self.word_index,
                                                 self.vector_dims,
                                                 self.embeddings_index)

    def layer(self, docs):

        '''Takes as input a series or array of documents and returns
        a Keras embeddings layer and x data. 

        USE:

        embeds_layer, x = e.layer(docs) 

        then in Keras model: 

        model.add(embeds_layer)
        model.fit(x=x)

        '''

        self.docs = docs

        self._matrix()

        embeddding_layer = Embedding(self.vocab_size,
                                     self.vector_dims,
                                     weights=[self.embedding_matrix],
                                     input_length=self.x.shape[1],
                                     trainable=False)

        return embeddding_layer, self.x



class Transform:

    def __init__(self, documents):

        '''Accepts input docs as list-of-lists where
        each sublist consist of a string value representing
        a single document, for example a paragraph.'''

        self._documents = documents

    def docs(self, clean=True, stopwords=True):

        '''Returns list-of-lists of strings'''

        from signs import Clean

        if clean:
            return [[Clean(doc, auto=True).text] for doc in self._documents]

        return self._documents

    def tokens(self, clean=True, stopwords=True):

        '''Returns a list-of-lists of tokens'''

        from signs.preprocess import create_tokens

        return create_tokens(self._documents, flatten=False, clean=clean)

    def tokens_flat(self, clean=True, stopwords=True):

        '''Returns a flattened list of tokens'''

        from signs.preprocess import create_tokens

        return create_tokens(self._documents, flatten=True, clean=clean)

    def docs_flat(self, clean=True, stopwords=True):

        '''Returns a flatted list of strings'''

        from signs.preprocess import lists_to_list

        return lists_to_list(self._documents)

    def docs_string(self, clean=True, stopwords=True):

        '''Returns a single string with blobbed documents'''

        from signs import Clean
        from signs.preprocess import lists_to_blob

        if clean:
            return lists_to_blob(
                [Clean(doc, auto=True).text for doc in self._documents])
        else:
            return lists_to_blob(self._documents)



from keras.preprocessing.text import Tokenizer
from keras.preprocessing.sequence import pad_sequences


def tokenize(texts, num_words=None, maxlen=12, padding='post'):

    '''TOKENIZE

    Create a word index and returns padded docs in return.

    '''

    tokenizer = Tokenizer(num_words=num_words)
    tokenizer.fit_on_texts(texts)
    encoded_docs = tokenizer.texts_to_sequences(texts)
    word_index = tokenizer.word_index

    x = pad_sequences(encoded_docs, maxlen=maxlen, padding=padding)
    vocab_size = max(list(word_index.values())) + 1

    return x, vocab_size, word_index



def similarity_docs(doc, model):

    '''Computes similarity scores for docs in the
    training corpus based on input doc.'''

    vecs = model.infer_vector(doc)
    sims = model.docvecs.most_similar([vecs],  topn=len(model.docvecs))

    return dict(sims)


def seen_similarity_matrix(model):

    '''Computes a similarity matrix for all the docs
    in the training corpus or in a set of documents
    based on the trained model.'''

    import numpy as np

    similarities = []
    out = []

    for i in range(len(model.docvecs)):

        # get similarities
        sims = model.docvecs.most_similar([model.docvecs[i]],
                                          topn=len(model.docvecs))

        # add to a list in dictionary format
        similarities.append(dict(sims))

    for sims in similarities:

        temp = []

        for i in range(len(sims)):
            temp.append(sims[i])

        out.append(temp)

    return np.array(out)


def unseen_similarity_matrix(model, docs):

    '''Computes a similarity matrix for unseen documents
    based on a trained doc2vec model.'''

    import numpy as np

    out = []

    for doc in docs:
        temp = []
        for i in range(len(docs)):
            sim = model.docvecs.similarity_unseen_docs(model, doc, docs[i])
            temp.append(sim)
        out.append(temp)

    return np.array(out)


def vector_spatial_distance(model, doc1, doc2):

    '''>> EXPERIMENTAL <<

    Computes spatial distance cosine for two documents
    based on a trained model.'''

    import scipy

    vec1 = model.infer_vector(doc1)
    vec2 = model.infer_vector(doc2)

    return 1 - scipy.spatial.distance.cosine(vec1, vec2)



class DocSimilarity:
    
    def __init__(self, model, docs):
        
        from signs.similarity import doc_similarity as sims
        from signs.utils.html_print import html_print
        
        self._sims = sims
        self._model = model
        self._docs = docs
        self._html_print = html_print
        
    def seen_matrix(self):
        
        '''creates a 2d matrix with similarities'''
    
        return self._sims.seen_similarity_matrix(self._model)

    def unseen_matrix(self, docs):
        
        '''same as above but for unseen docs'''
        
        return self._sims.unseen_similarity_matrix(self._model, docs)

    def similar_docs(self, doc):
    
        '''for comparing a single doc against all seen docs'''
        return self._sims.similarity_docs(doc, self._model)

    def spatial_distance(self, doc1, doc2):
        
        '''for comparing two unseen or seen docs'''
        return self._sims.vector_spatial_distance(self._model,
                                                  doc1,
                                                  doc2)
    
    def preview_results(self, docs=None):
        
        if docs is None:
            # get the keys from docs with similarities
            similarities = self._get_similarities(self._docs.docs())
        else:
            similarities = self._get_similarities(docs)
        
        # print out the highest and lowest match
        self._print_highest(similarities)
        self._print_lowest(similarities)
    

    def _get_similarities(self, docs):

        import random

        # pick a ramdom document
        doc_id = random.randint(0, len(docs))

        # find similar documents
        similarities = self._sims.similarity_docs(docs[doc_id], self._model)

        return similarities

    def _print_highest(self, similarities):

        # create content
        text = self._docs.docs(False)[list(similarities.keys())[0]][0]

        # create similarity value and round
        similarity = similarities[list(similarities.keys())[0]]
        similarity = round(similarity, 4)

        # parse together the title
        title = "HIGHEST MATCH : " + str(similarity)

        # print it out
        self._html_print(text, title)


    def _print_lowest(self, similarities):

        # create content
        text = self._docs.docs(False)[list(similarities.keys())[-1]][0]

        # create similarity value and round
        similarity = similarities[list(similarities.keys())[-1]]
        similarity = round(similarity, 4)

        # parse together the title
        title = "LOWEST MATCH : " + str(similarity)

        # print it out
        self._html_print(text, title)